import EventEmitter from "eventemitter3";
import * as Automerge from "@automerge/automerge";
import debug from "debug";
import { decodeChange } from "@automerge/automerge";
const log = debug("DocSynchronizer");
const conciseLog = debug("DocSynchronizer:Concise"); // Only logs one line per receive/send
const opsLog = debug("DocSynchronizer:Ops"); // Log list of ops of each message
/**
 * DocSynchronizer takes a handle to an Automerge document, and receives & dispatches sync messages
 * to bring it inline with all other peers' versions.
 */
export class DocSynchronizer extends EventEmitter {
    handle;
    // we track peers separately from syncStates because we might have more syncStates than active peers
    peers = [];
    syncStates = {}; // peer -> syncState
    constructor(handle) {
        super();
        this.handle = handle;
        handle.on("change", () => this.syncWithPeers());
    }
    getSyncState(peerId) {
        if (!peerId) {
            throw new Error("Tried to load a missing peerId");
        }
        if (!this.peers.includes(peerId)) {
            log("adding a new peer", peerId);
            this.peers.push(peerId);
        }
        let syncState = this.syncStates[peerId];
        if (!syncState) {
            syncState = Automerge.initSyncState();
        }
        return syncState;
    }
    setSyncState(peerId, syncState) {
        this.syncStates[peerId] = syncState;
    }
    async sendSyncMessage(peerId, documentId, doc) {
        // TODO: Right now this can send duplicate messages
        // even if nothing has changed
        log(`[${this.handle.documentId}]->[${peerId}]: sendSyncMessage`);
        const syncState = this.getSyncState(peerId);
        const [newSyncState, message] = Automerge.generateSyncMessage(doc, syncState);
        this.setSyncState(peerId, newSyncState);
        if (message) {
            const decoded = Automerge.decodeSyncMessage(message);
            log(`[${this.handle.documentId}]->[${peerId}]: sendSyncMessage: ${message.byteLength}b`, decoded);
            conciseLog(`${peerId} ⬅️️  ${this.handle.documentId} (${message.byteLength} bytes)`);
            if (opsLog.enabled) { // guard opsLog, so decodeChange is not called unnecessarily, because it can be expensive
                opsLog(decoded.changes.flatMap(change => {
                    return decodeChange(change).ops.map((op) => JSON.stringify(op));
                }));
            }
            const channelId = this.handle.documentId;
            this.emit("message", {
                targetId: peerId,
                channelId,
                message,
                broadcast: false,
            });
        }
        else {
            log(`[${this.handle.documentId}]->[${peerId}]: sendSyncMessage: [no message generated]`);
        }
    }
    async beginSync(peerId) {
        log(`[${this.handle.documentId}]: beginSync: ${peerId}`);
        const { documentId } = this.handle;
        const doc = await this.handle.syncValue();
        // Just in case we have a sync state already, we round-trip it through
        // the encoding system to make sure state is preserved. This prevents an
        // infinite loop caused by failed attempts to send messages during disconnection.
        // TODO: we should be storing sync states and besides, we only need to do this on reconnect
        this.setSyncState(peerId, Automerge.decodeSyncState(Automerge.encodeSyncState(this.getSyncState(peerId))));
        this.sendSyncMessage(peerId, documentId, doc);
    }
    endSync(peerId) {
        log(`removing peer ${peerId}`);
        this.peers = this.peers.filter((p) => p !== peerId);
    }
    async onSyncMessage(peerId, channelId, message) {
        if (channelId !== this.handle.documentId) {
            throw new Error(`[DocHandle: ${this.handle.documentId}]: Received a sync message for ${channelId}`);
        }
        // We need to block receiving the syncMessages until we've checked local storage
        // TODO: this is kind of an opaque way of doing this...
        // await this.handle.syncValue()
        this.handle.updateDoc((doc) => {
            const decoded = Automerge.decodeSyncMessage(message);
            log(`[${this.handle.documentId}]->[${peerId}]: receiveSync: ${message.byteLength}b`, decoded);
            conciseLog(`${peerId} ➡️️️  ${this.handle.documentId} (${message.byteLength} bytes)`);
            if (opsLog.enabled) { // guard opsLog, so decodeChange is not called unnecessarily, because it can be expensive
                opsLog(decoded.changes.flatMap(change => {
                    return decodeChange(change).ops.map((op) => JSON.stringify(op));
                }));
            }
            const start = Date.now();
            const [newDoc, newSyncState] = Automerge.receiveSyncMessage(doc, this.getSyncState(peerId), message);
            const end = Date.now();
            const time = end - start;
            log(`[${this.handle.documentId}]: receiveSync: <- ${peerId} ${message.byteLength}b in ${time}ms ${time > 1000 ? "[SLOW]!" : ""} from ${peerId}`);
            this.setSyncState(peerId, newSyncState);
            // respond to just this peer (as required)
            this.sendSyncMessage(peerId, this.handle.documentId, doc);
            return newDoc;
        });
    }
    async syncWithPeers() {
        log(`[${this.handle.documentId}]: syncWithPeers`);
        const { documentId } = this.handle;
        const doc = await this.handle.syncValue();
        this.peers.forEach((peerId) => {
            this.sendSyncMessage(peerId, documentId, doc);
        });
    }
}
